"""ConsumerParallel. Acts as an interface between Akita and the ETL pipeline."""
import logging
import time
from collections import OrderedDict
from typing import Any

from dask.distributed import Client, Future  # , as_completed

from .core import EventConsumer

logger = logging.getLogger(__name__)


class ConsumerParallel(EventConsumer):
    """
    ConsumerParallel that runs on a thread as a daemon process.

    This event consumer produce futures that are consumed in parallel
        by multiple Dask workers.
    Each future corresponds to one or multiple jobs.

    Attributes
    ----------
    _queue
        Queue where events are placed.
    _job_producer
        Producer of jobs.
    _worker, optional
        Worker thread that consumes events from the and processes them to produce jobs.
    _client
        Dask client.
    _workers_per_event
        Number of Dask workers per future.
    _futures
        List of Dask futures.
    _max_futures
        Maximum number of simultaneous Dask futures.
    """

    def __init__(
        self, cluster_client: Client, workers_per_event: int = 1, *args, **kwargs
    ) -> None:
        """
        Initialize the event consumer.

        Parameters
        ----------
        client
            Dask client.
        args
            Arguments to pass to EventConsumer.__init__.
        kwargs
            Keyword arguments to pass to EventConsumer.__init__.
        """
        super().__init__(*args, **kwargs)

        self._client = cluster_client
        self._workers_per_event = workers_per_event
        self._max_futures: int = None
        self._futures: OrderedDict[str, Any] = None

    def _purge_future(self, future: Future) -> None:
        """Purge a future.

        Parameters
        ----------
        future
            Future that has succeeded.
        """
        del self._futures[future], future

    def _submit_future(self, event: Any) -> Future:
        """
        Submit a future to the Dask cluster.

        Parameters
        ----------
        event
            Event to be submitted.

        Returns
        -------
        Future
        """
        logger.info(f"Submitting future for: {event}")

        # Submit the future to the Dask cluster
        future = self._client.submit(self._job_producer.produce_jobs, event)

        # Add the future to the list of futures and to the events dictionary
        self._futures[future] = event

        return future

    def _create_futures(self) -> None:
        """Create futures."""
        # Get the events from the queue for which a future has not yet been created
        if self._futures is None:
            self._futures = OrderedDict()

        diff = set(self._queue.queue_list) - set(self._futures.values())
        for event in diff:
            if len(self._futures) < self._max_futures:
                self._submit_future(event)

    def _process_succeeded_future(self, future: Future) -> None:
        """
        Process a future that has succeeded.

        Parameters
        ----------
        future
            Future that has succeeded.
        """
        event = self._futures[future]
        logger.info(f"Event consumed: {event}")

        # Remove event from the queue and purge the future
        self._queue.remove(event)
        self._purge_future(future)

    def _process_failed_future(self, future: Future) -> None:
        """
        Process a future that has failed.

        Parameters
        ----------
        future
            Future that has failed.
        """
        event = self._futures[future]
        logger.warning(
            f"Event {event} finished with an error: {future.exception()}."
            + " Future will be retried.."
        )
        # Wait 1 second before retrying the future
        time.sleep(1)
        future.retry()

    def _process_finished_future(self, future: Future) -> None:
        """
        Process a future that has finished.

        Parameters
        ----------
        future
            Future that has finished.

        Raises
        ------
        ValueError
            If the status of the future is not "finished" or "error".
        """
        if future.status == "finished":
            self._process_succeeded_future(future)
        elif future.status == "error":
            self._process_failed_future(future)
        elif future.status == "pending":
            pass
        else:
            raise ValueError(f"Invalid future status: {future.status}")

    def _process_futures(self) -> None:
        """Process futures as they finish."""
        # Alternative loop to process futures
        # completed_futures = as_completed(
        #     self._futures, with_results=True, raise_errors=False
        # )
        # for future, result in completed_futures:
        #     self._process_finished_future(future)

        for future in self._futures.copy():
            self._process_finished_future(future)

    def _run_event_loop(self, sleep_time: int = 5) -> None:
        """
        Run the worker thread.

        Callback function to be used as a target by Thread.

        Parameters
        ----------
        sleep_time
            Sleep time in seconds for which the thread is idle.
        """
        while True:
            if self._queue.get_queue_size():
                # Calculate maximum number of futures given the
                # number of workers currently running
                self._max_futures = (
                    len(self._client.scheduler_info()["workers"])
                    // self._workers_per_event
                )

                # Create futures if there are events in the queue
                # that have not been 'futurized' yet and if the current number of futures
                # is less than the maximum number of futures
                self._create_futures()

                # Process the futures when they are finished
                self._process_futures()

                time.sleep(sleep_time)
            elif self._is_sentinel_active():
                logger.info("The queue is empty and got an end-of-queue sentinel")
                logger.info("The event consumer is exiting...")
                break
            else:
                time.sleep(sleep_time)



"""pyyaml constructors."""
import importlib

import yaml
from dask.distributed import Client

from ..akita.core import Akita
from ..akita.factory import get_akita_dependencies
from ..etl_pipeline.basic_pipeline import BasicPipeline
from ..etl_pipeline.celery_pipeline import CeleryPipeline
from ..etl_pipeline.core import Job, Task
from ..event_consumer.consumer_parallel import ConsumerParallel
from ..event_consumer.consumer_serial import ConsumerSerial
from ..filesystems.object_store import ObjectStoreS3


def dask_client_constructor(loader: yaml.SafeLoader, node: yaml.MappingNode) -> Client:
    """Create a Dask client bound to a cluster ."""
    params = loader.construct_mapping(node, deep=True)
    module_name, func_name = params["cluster"].rsplit(".", 1)
    module = importlib.import_module(module_name)

    cluster = getattr(module, func_name)
    kwargs = {
        str(key): params[key] for key in params if key not in ["cluster", "scale"]
    }
    cluster = cluster(**kwargs)

    if "scale" in params:
        if isinstance(params["scale"], dict):
            cluster.scale(**params["scale"])
        elif isinstance(params["scale"], list):
            cluster.scale(*params["scale"])
        else:
            cluster.scale(params["scale"])

    return Client(cluster)


def object_constructor(loader: yaml.SafeLoader, node: yaml.MappingNode) -> object:
    """Create an object."""
    params = loader.construct_mapping(node, deep=True)
    kwargs = {str(key): params[key] for key in params if key != "class"}

    params = loader.construct_mapping(node, deep=True)
    module_name, class_name = params["class"].rsplit(".", 1)
    module = importlib.import_module(module_name)

    return getattr(module, class_name)(**kwargs)


def object_store_constructor(
    loader: yaml.SafeLoader, node: yaml.MappingNode
) -> ObjectStoreS3:
    """Construct an Object Store S3 instance."""
    kwargs = {
        str(key): val for key, val in loader.construct_mapping(node, deep=True).items()
    }
    return ObjectStoreS3(**kwargs)


def akita_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> Akita:
    """Construct an Akita instance."""
    params = loader.construct_mapping(node, deep=True)
    kwargs = {str(key): params[key] for key in params if key not in ["monitor"]}
    dependencies = get_akita_dependencies(**kwargs)
    return Akita(*dependencies, monitor=params["monitor"])


def consumer_serial_constructor(
    loader: yaml.SafeLoader, node: yaml.nodes.MappingNode
) -> ConsumerSerial:
    """Construct a ConsumerSerial instance."""
    params = loader.construct_mapping(node, deep=True)
    kwargs = {str(key): params[key] for key in params if key not in ["akita"]}
    akita = params["akita"]

    return ConsumerSerial(queue=akita.queue, **kwargs)


def consumer_parallel_constructor(
    loader: yaml.SafeLoader, node: yaml.nodes.MappingNode
) -> ConsumerParallel:
    """Construct a ConsumerParallel instance."""
    params = loader.construct_mapping(node, deep=True)
    kwargs = {str(key): params[key] for key in params if key not in ["akita"]}
    return ConsumerParallel(queue=params["akita"].queue, **kwargs)


def celery_pipeline_constructor(
    loader: yaml.SafeLoader, node: yaml.nodes.MappingNode
) -> CeleryPipeline:
    """Construct a CeleryPipeline instance."""
    kwargs = {
        str(key): val for key, val in loader.construct_mapping(node, deep=True).items()
    }
    return CeleryPipeline(**kwargs)


def basic_pipeline_constructor(
    loader: yaml.SafeLoader, node: yaml.nodes.MappingNode
) -> BasicPipeline:
    """Construct a BasicPipeline instance."""
    kwargs = {
        str(key): val for key, val in loader.construct_mapping(node, deep=True).items()
    }
    return BasicPipeline(**kwargs)


def job_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> Job:
    """Construct a Job."""
    kwargs = {
        str(key): val for key, val in loader.construct_mapping(node, deep=True).items()
    }
    return Job(**kwargs)


def task_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> Task:
    """Construct a Job."""
    params = loader.construct_mapping(node, deep=True)
    kwargs = {key: params[key] for key in params if key != "function"}

    params = loader.construct_mapping(node, deep=True)
    module_name, func_name = params["function"].rsplit(".", 1)
    module = importlib.import_module(module_name)

    return Task(function=getattr(module, func_name), kwargs=kwargs)


constructors_dict = {
    "!Job": job_constructor,
    "!Task": task_constructor,
    "!Akita": akita_constructor,
    "!ConsumerSerial": consumer_serial_constructor,
    "!ConsumerParallel": consumer_parallel_constructor,
    "!BasicPipeline": basic_pipeline_constructor,
    "!CeleryPipeline": celery_pipeline_constructor,
    "!DaskClient": dask_client_constructor,
    "!ObjectStoreS3": object_store_constructor,
}



"""pyyaml loader."""
from typing import Any

import yaml

from .yaml_constructors import constructors_dict
from .yaml_executers import executers_dict


def get_loader():
    """Add constructors and executors to PyYAML loader."""
    loader = yaml.UnsafeLoader

    for tag, constructor in constructors_dict.items():
        loader.add_constructor(tag, constructor)

    for tag, executer in executers_dict.items():
        loader.add_constructor(tag, executer)

    return loader


def load_yaml(yaml_file: str, loader=yaml.SafeLoader) -> Any:
    """
    Load an yaml yaml file.

    Parameters
    ----------
    yaml_file
        Path to the yaml file.
    loader
        PyYAML loader.

    Returns
    -------
        yaml file content.
    """
    with open(yaml_file, "r") as stream:
        try:
            yaml_contents = yaml.load(stream, Loader=loader)
        except yaml.YAMLError as exc:
            raise yaml.YAMLError(exc)

    return yaml_contents



"""Argument parser module."""
import argparse

from ..__init__ import __version__


def create_parser():
    """Create the argument parser."""
    parser = argparse.ArgumentParser(
        description=f"dpypeline {__version__} command line interface",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-i",
        "--input",
        dest="input_file",
        default="pipeline.yaml",
        help="filepath to the pipeline YAML file",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )

    return parser



"""dpypeline command line interface."""
import logging
import sys
import time

from .argument_parser import __version__, create_parser
from .yaml_loader import get_loader, load_yaml

logger = logging.getLogger(__name__)


def banner():
    """Log the dpypeline banner."""
    logger.info(
        r"""
      _                             _  _
     | |                           | |(_)
   __| | _ __   _   _  _ __    ___ | | _  _ __    ___
  / _` || '_ \ | | | || '_ \  / _ \| || || '_ \  / _ \
 | (_| || |_) || |_| || |_) ||  __/| || || | | ||  __/
  \__,_|| .__/  \__, || .__/  \___||_||_||_| |_| \___|
        | |      __/ || |
        |_|     |___/ |_|
""",
        extra={"simple": True},
    )


def start_banner():
    """Log the dpypeline start banner."""
    banner()
    logger.info(f"version: {__version__}", extra={"simple": True})


def exit_banner():
    """Log the dpypeline exit banner."""
    banner()
    logger.info("dpypeline has terminated succesfully! :)", extra={"simple": True})


def dpypeline():
    """Run the dpypeline."""
    logging.basicConfig(
        format="dpypeline | %(levelname)s | %(asctime)s | %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )
    start_banner()

    parser = create_parser()
    args = parser.parse_args()

    settings = load_yaml(args.input_file, loader=get_loader())

    akita = settings["akita"]
    event_consumer = settings["event_consumer"]
    settings["pipeline"]

    akita.run()
    event_consumer.run()

    if "dask_client" in settings:
        # Close the Dask client if it exists.
        logger.info("Closing Dask client...")
        settings["dask_client"].shutdown()
        settings["dask_client"].close()
        logger.info("Dask client has been closed successfully.")
        time.sleep(1)

    exit_banner()



"""Basic pipeline module."""
import logging
from typing import Any

from .core import ETLPipeline, Job

logger = logging.getLogger(__name__)


class BasicPipeline(ETLPipeline):
    """Basic pipeline class."""

    def __init__(self, jobs: list[Job] = None) -> None:
        """Initialise the pipeline."""
        super().__init__(jobs)

    def produce_jobs(self, event: Any) -> list[Any]:
        """
        Produce jobs to be run sequentially.

        Parameters
        ----------
        event
            Triggering event.

        Returns
        -------
            List of results of the jobs.
        """
        logger.debug(f"Producing jobs for event {event}.")
        results = [job.run(event) for job in self.jobs]
        logger.debug(f"Jobs for event {event} have been produced successfully.")

        return results



"""S3 object store class."""
import io
import json
import logging
import os
from typing import IO

import fsspec
import numpy as np
import s3fs


class ObjectStoreS3(s3fs.S3FileSystem):
    """
    S3 object store.

    Parameters
    ----------
    s3fs
        _description_
    """

    def __init__(
        self,
        anon: bool = False,
        store_credentials_json: str | None = None,
        secret: str | None = None,
        key: str | None = None,
        endpoint_url: str | None = None,
        *fs_args,
        **fs_kwargs,
    ) -> None:
        """
        Initialize the S3 object store.

        Parameters
        ----------
        anon, optional
            _description_, by default False
        store_credentials_json, optional
            _description_, by default None
        secret, optional
            _description_, by default None
        key, optional
            _description_, by default None
        endpoint_url, optional
            _description_, by default None
        """
        self._anon = anon
        logging.info("-" * 79)
        if store_credentials_json is None:
            logging.info(
                "No JSON file was provided."
                "Object store credentials will be obtained from the arguments passed."
            )
            self._store_credentials = {
                "secret": secret,
                "token": key,
                "endpoint_url": endpoint_url,
            }
        else:
            logging.info(
                "Object store credentials will be read from the JSON file "
                + f"{store_credentials_json}"
            )
            self._store_credentials = self.load_store_credentials(
                store_credentials_json
            )

        self._remote_options = self.get_remote_options(override=True)

        super().__init__(*fs_args, **self._remote_options, **fs_kwargs)

    @staticmethod
    def load_store_credentials(path: str) -> dict:
        """
        Set the credentials of the object store from a JSON file.

        Parameters
        ----------
        path
            Absolute or relative filepath to the JSON file containing
            the object store credentials.

        Returns
        -------
        store_credentials
            Dictionary containing the values of the `token`,
                `secret` and `endpoint_url` keys used
            to access the object store.
        """
        try:
            with open(path) as f:
                store_credentials = json.load(f)
        except Exception as error:
            raise Exception(error)

        for key in ["token", "secret", "endpoint_url"]:
            if key not in store_credentials:
                logging.info("-" * 79)
                logging.warning(
                    f'"{key}" is not a key in the JSON file provided. '
                    + "Its value will be set to None."
                )

        return store_credentials

    def create_bucket(self, bucket: str, **kwargs) -> None:
        """
        Create a bucket in the object store.

        Parameters
        ----------
        bucket
            Bucket to create.
        """
        try:
            return self.mkdir(bucket, **kwargs)
        except FileExistsError:
            logging.info(f"Bucket '{bucket}' already exists.")

    def get_remote_options(self, override: bool = False) -> dict:
        """
        Get the remote options of the object store.

        Parameters
        ----------
        override
            Flag to create remote_options from scratch (True)
            or to simply retrieve the current dict (False).

        Returns
        -------
        remote_options
            Dictionary containing the remote options of the object store.

        """
        if override:
            self._remote_options = {
                "anon": self._anon,
                "secret": self._store_credentials["secret"],
                "key": self._store_credentials["token"],
                "client_kwargs": {
                    "endpoint_url": self._store_credentials["endpoint_url"]
                },
            }

        return self._remote_options

    def get_mapper(
        self, bucket: str, prefix: str = "s3://", **get_mapper_kwargs
    ) -> fsspec.mapping.FSMap:
        """
        Make a MutableMaping interface to the desired bucket.

        Parameters
        ----------
        bucket
            Name of the bucket to place the file in.
        prefix: str, default "s3://"
            Protocol prefix
        **get_mapper_kwargs
            Kwargs for get_mapper. See: https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.get_mapper.

        Returns
        -------
        mapper
            Dict-like key-value store.
        """  # noqa adamwa
        mapper = fsspec.get_mapper(
            prefix + bucket, **self._remote_options, **get_mapper_kwargs
        )

        return mapper

    def get_bucket_list(self) -> list[str]:
        """
        Get the list of buckets in the object store.

        Returns
        -------
        bucket_list
            List of the object store buckets.
        """
        return self.ls("/")

    def write_file_to_bucket(
        self,
        path: str | os.PathLike | IO,
        bucket: str,
        file_name: str,
        chunk_size: int = -1,
        parallel: bool = False,
    ) -> None:
        """
        Write to a bucket of the object store.

        Parameters
        ----------
        path
            Absolute or relative filepath of the file to be written to the object store,
            or file-like object to be written to the object store.
        bucket
            Name of the bucket to place the file in.
        file_name
            Name that will be used to identify the file in the bucket.
        chunk_size
            Size of the chunk (in bytes) to read/write at once.
            If the chunk size is -1, the file will be read/written at once.
        parallel
            Flag to enable parallel writing (True) or not (False).
        """
        assert (
            bucket.split(os.path.sep, 1)[0] in self.get_bucket_list()
        ), f'Bucket "{bucket}" does not exist.'

        if isinstance(path, str) or isinstance(path, os.PathLike):
            assert os.path.isfile(path), f'"{path}" is not a file.'
            file_size = os.path.getsize(path)
        elif isinstance(path, io.IOBase):
            file_size = path.seek(0, os.SEEK_END)
        else:
            raise ValueError(f'"{path}" is not file-like, path-like or a string.')

        # create the empty file
        dest_path = os.path.join(bucket, file_name)
        self.open(dest_path, mode="wb", s3=dict(profile="default")).close()

        # create the chunks offsets and lengths
        chks = self._create_chunks_offsets_lengths(file_size, chunk_size)

        # write the file to the bucket
        self._write_to_bucket(path, dest_path, chks, parallel)

    def _write_to_bucket(
        self,
        path: str | os.PathLike | IO,
        dest_path: str,
        chunks_offsets_lengths: dict,
        parallel: bool,
    ) -> None:
        """
        Write to a bucket of the object store.

        Parameters
        ----------
        path
            Absolute or relative filepath of the file to be written to the object store,
            or file-like object to be written to the object store.
        dest_path
            Name of the bucket to place the file in.
        chunks_offsets_lengths
            Dictionary containing the chunk offsets and lengths
            of the file to be written to the object store.
        parallel
            Flag to enable parallel writing (True) or not (False).
        """
        n_chunks = len(chunks_offsets_lengths)

        # Arguments to be fed to the map
        multi_part_files = [f"{dest_path}.part{i}" for i in range(n_chunks)]
        chks_offsets = [
            chunks_offsets_lengths[chk]["offset"] for chk in chunks_offsets_lengths
        ]
        chks_lengths = [
            chunks_offsets_lengths[chk]["length"] for chk in chunks_offsets_lengths
        ]
        paths = [path] * n_chunks

        if parallel:
            import dask

            dask.compute(
                list(
                    map(
                        dask.delayed(self._write_chunk),
                        paths,
                        multi_part_files,
                        chks_offsets,
                        chks_lengths,
                    )
                )
            )
        else:
            list(
                map(
                    self._write_chunk,
                    paths,
                    multi_part_files,
                    chks_offsets,
                    chks_lengths,
                )
            )

        if n_chunks > 1:
            # Remove any partial uploads in the bucket associated with the file
            # Create single S3 file from list of S3 files
            self.merge(dest_path, multi_part_files)
            # Remove any partial uploads in the bucket associated with the file
            self.rm(multi_part_files)

    def __open(self, path: str | os.PathLike | IO, mode: str = "rb") -> IO:
        """
        Open a file using a bespoke interface.

        Parameters
        ----------
        path
            Absolute or relative filepath of the file to be opened,
            or file-like object to be opened.
        mode
            File open mode.

        Returns
        -------
        filelike object
        """
        if isinstance(path, str) or isinstance(path, os.PathLike):
            return open(path, mode)
        elif isinstance(path, io.IOBase):
            # filelike object
            return path
        else:
            raise ValueError(f"Unsupported file type: {type(path)}")

    def _write_chunk(
        self, path_read, path_write, chunk_offset: int, chunk_length: int
    ) -> None:
        """
        Write a chunk to an open file.

        Parameters
        ----------
        path_read
            File to read from.
        path_write
            File to write to.
        chunk_offset
            Offset of the chunk (in bytes).
        chunk_length
            Length of the chunk (in bytes).
        """
        # Read the chunk
        with self.__open(path_read, mode="rb") as f:
            f.seek(chunk_offset, 0)
            bytechunk = f.read(chunk_length)

        # Write the chunk
        with self.open(path_write, mode="wb", s3=dict(profile="default")) as f:
            f.write(bytechunk)

    def _create_chunks_offsets_lengths(self, nbytes, chunk_size: int) -> dict:
        """
        Create the chunks offsets and lengths for a file with nbytes given the chunk_size.

        Parameters
        ----------
        nbytes
            Number of bytes of the file.
        chunk_size
            Size of the chunk (in bytes).

        Returns
        -------
        chunks_offsets_lengths
            Dictionary containing the chunks offsets and lengths.
        """
        chunks_offsets_lengths = {}

        if chunk_size == -1:
            chunk_size = nbytes

        # Calculate the number of chunks
        nchunks = int(np.ceil(nbytes / chunk_size))

        # Calculate the chunks offsets and length
        for i in range(nchunks):
            chunks_offsets_lengths[i] = {
                "offset": i * chunk_size,
                "length": chunk_size,
            }

        return chunks_offsets_lengths



"""ConsumerSerial. Acts as an interface between Akita and the ETL pipeline."""
import logging
import time

from .core import EventConsumer

logger = logging.getLogger(__name__)


class ConsumerSerial(EventConsumer):
    """
    ConsumerSerial that runs on a thread as a daemon process.

    This event consumer consumes events from an in-memory queue and processes
    them to produce jobs that are run in serial.
    """

    def _consume_event(self, event) -> None:
        """
        Consume event by producing jobs.

        Parameters
        ----------
        event
            Event representing file or directory creation, deletion,
            modification or moving.
        """
        logger.info(f"Consuming event '{event}'...")
        self._job_producer.produce_jobs(event=event)
        logger.info(f"Event '{event}' has been consumed successfully.")

    def _run_event_loop(self, sleep_time: int = 5) -> None:
        """
        Run the worker thread.

        Callback function to be used as a target by Thread.

        Parameters
        ----------
        sleep_time
            Sleep time in seconds for which the thread is idle.
        """
        while True:
            if self._queue.get_queue_size():
                event = self._queue.peek()
                logger.info("-" * 79)
                self._consume_event(event)
                self._queue.dequeue()
            elif self._is_sentinel_active():
                logger.info("The queue is empty and got an end-of-queue sentinel")
                logger.info("The event consumer is exiting...")
                break
            else:
                time.sleep(sleep_time)



"""pyyaml exectures."""
import importlib


def function_executer(loader, node):
    """Execute a function."""
    params = loader.construct_mapping(node, deep=True)
    module_name, func_name = params["function"].rsplit(".", 1)
    module = importlib.import_module(module_name)
    func = getattr(module, func_name)
    kwargs = {key: params[key] for key in params if key != "function"}
    return func(**kwargs)


def method_executer(loader, node):
    """Execute a method."""
    params = loader.construct_mapping(node, deep=True)
    instance = params["instance"]
    method = getattr(instance, params["method"])
    kwargs = {key: params[key] for key in params if key not in ["instance", "method"]}
    return method(**kwargs)


executers_dict = {
    "!Function": function_executer,
    "!Method": method_executer,
}



"""Queue of events class."""
from __future__ import annotations

import logging
import os
import pickle
from queue import Empty, Full, Queue
from typing import Any

logger = logging.getLogger(__name__)


class Sentinel:
    """A Sentinel object to sinalise whether to close the queue when it is empty."""

    def __init__(self, active: bool = False) -> None:
        """Initialize the Sentinel object."""
        self._active = active

    def set_state(self, active: bool) -> bool:
        """Set the state of the sentinel."""
        self._active = active
        return self._active

    def __call__(self, *args: Any, **kwargs: Any) -> bool:
        """Return the state of the sentinel."""
        return self._active


class EventsQueue(Queue):
    """EventsQueue singleton class."""

    _instance: EventsQueue = None
    _initialized: bool = False
    _state_file_suffix: str = "queue_state.pickle"
    _processed_events_file_suffix: str = "processed_events.pickle"

    def __init__(self, maxsize: int = 0) -> None:
        """Initiate the EventsQueue singleton."""
        assert (
            os.getenv("CACHE_DIR") is not None
        ), "CACHE_DIR environmental variable is not set."

        if not self._initialized:
            logger.debug("Initializing singleton instance of EventsQueue.")
            super().__init__(maxsize=maxsize)
            self._initialized: bool = True
            self._processed_events: list[str] = None
            self._state_file = os.path.join(
                os.getenv("CACHE_DIR"), self._state_file_suffix
            )
            self._load_state()

        self._processed_events_file = str = os.path.join(
            os.getenv("CACHE_DIR"), self._processed_events_file_suffix
        )

        self._sentinel = Sentinel()

    def __new__(cls, maxsize: int = 0) -> EventsQueue:
        """
        Create a new instance of the EventsQueue class using the Singleton pattern.

        Returns
        -------
            Instance of the EventsQueue class.
        """
        if cls._instance is None:
            logger.debug("Creating singleton instance of EventsQueue.")
            cls._instance = super(EventsQueue, cls).__new__(cls)
            logger.debug(
                "Creation of singleton instance of EventsQueue has been succesful."
            )

        return cls._instance

    @property
    def queue_list(self) -> list[Any]:
        """
        Return the list of events in the queue.

        Returns
        -------
            List of events in the queue.
        """
        return list(self.queue)

    @property
    def sentinel(self) -> Any:
        """Return the sentinel."""
        return self._sentinel

    def set_sentinel_state(self, active: bool) -> bool:
        """
        Set the sentinel state.

        Parameters
        ----------
        active
            The new state of the sentinel.
        """
        return self._sentinel.set_state(active)

    @property
    def processed_events(self) -> list[Any]:
        """
        Return the list of events processed thus far.

        Returns
        -------
            List of events processed thus far.
        """
        if self._processed_events is None:
            # Only load processed events if first time called.
            self._load_processed_events()

        return self._processed_events

    def _save_state(self, logger_prefix="") -> None:
        """Save the state of the queue."""
        logger.debug(f"{logger_prefix}Saving state of the queue.")

        with open(self._state_file, "wb") as f:
            pickle.dump(self.queue, f)

        logger.debug(
            f"{logger_prefix}Save of the state of the queue has been saved succesful."
        )

    def _load_state(self) -> None:
        """Load the state of the queue."""
        logger.debug("Loading state of the queue.")

        if os.path.isfile(self._state_file):
            logger.debug(f"Found queue state file {self._state_file}")

            with open(self._state_file, "rb") as f:
                self.queue = pickle.load(f)
        else:
            logger.debug(f"No queue state file {self._state_file} was found.")

    def _load_processed_events(self) -> None:
        """Load the processed events."""
        logger.debug("Loading processed events.")

        if os.path.isfile(self._processed_events_file):
            logger.debug(f"Found processed events file {self._processed_events_file}.")

            with open(self._processed_events_file, "rb") as f:
                self._processed_events = pickle.load(f)
        else:
            logger.debug(
                f"No processed events file {self._processed_events_file} was found."
            )
            self._processed_events = []

    def _save_processed_events(self) -> None:
        """Save the events processed so far in this session."""
        logger.debug("Saving processed events.")
        with open(self._processed_events_file, "wb") as f:
            pickle.dump(self._processed_events, f)
        logger.debug("Save of processed events has been successful.")

    def enqueue(self, event: Any) -> bool:
        """Add an event to the queue.

        Everytime an event is added to the queue, the state of the queue is saved.

        Parameters
        ----------
        event
            Event to add to the queue.

        Returns
        -------
        True if the event was added to the queue.
        """
        try:
            logger.debug(f"Enqueuing event '{event}'.")
            self.put(event, block=False)
            self._save_state(logger_prefix="Enqueueing event: ")
            logger.debug(f"Event '{event}' has been enqueued succesfully.")
            return True
        except Full:
            raise Full("Queue is full.")

    def _set_as_processed(self, event: Any) -> None:
        """Set the event as processed."""
        # Set as processesed event
        if self._processed_events is None:
            self._processed_events = []

        self._processed_events.append(event)
        self._save_processed_events()

    def dequeue(self) -> Any:
        """
        Remove and return an event from the queue.

        If the queue is empty, returns None.

        Returns
        -------
        First event in the queue or None if the queue is empty.
        """
        try:
            logger.debug("-" * 79)
            logger.debug("Dequeuing event: Starting.")
            event = self.get(block=False)
            logger.debug(f"Dequeued event: {event}.")
            self._save_state(logger_prefix="Dequeuing event: ")
            self._set_as_processed(event)
            return event
        except Empty:
            return None

    def peek(self) -> Any:
        """
        Peek first item in the queue withou removing it.

        If the queue is empty, returns None.

        Returns
        -------
        First event in the queue or None if the queue is empty.
        """
        if self.get_queue_size():
            logger.debug("-" * 79)
            logger.debug("Peeking first item in the queue.")
            return self.queue[0]

    def get_queue_size(self) -> int:
        """
        Get the size of the queue.

        Returns
        -------
        Number of events in the queue.
        """
        return self.qsize()

    def remove(self, event: Any) -> int:
        """
        Remove an event from the queue.

        Notes
        -----
        Removing elments from arbitrary positions in the queue should be avoided,
        as it goes against the FIFO principle.
        However, this method is useful for removing events processed by
        parallel workers.

        Parameters
        ----------
        event
            Event to add to the queue.

        Returns
        -------
            True if the event was removed from the queue, False otherwise.
        """
        try:
            logger.debug("-" * 79)
            logger.debug(f"Removing event: {event}.")
            self.queue.remove(event)
            self._save_state(logger_prefix="Removing event: ")
            self._set_as_processed(event)
            return True
        except ValueError as e:
            logger.error(f"{e}")
            return False

    @classmethod
    def clear_instance(cls):
        """Clear the singleton instance."""
        cls._instance = None



"""Test suite for the ETLPipeline core."""
from dpypeline.etl_pipeline.core import ETLPipeline, Job, Task


def test_create_task() -> None:
    """Test the Task class."""
    function = min
    task = Task(function=function)
    assert isinstance(task, Task)
    assert task.function == function
    assert task.args == ()
    assert task.kwargs == {}

    args = ([1, 2, 3],)
    task = Task(function=sum, args=args)
    assert task.function == sum
    assert task.args == args
    assert task.kwargs == {}
    assert task.run() == 6

    task = Task(function=sum)
    assert task.function == sum
    assert task.args == ()
    assert task.kwargs == {}
    assert task.run([3, 4, 7]) == 14

    kwargs = {"a": 1, "b": 2, "c": 3}

    def dummy_func(a, b, c):
        return a + b + c

    task = Task(function=dummy_func, kwargs=kwargs)
    assert task.function == dummy_func
    assert task.args == ()
    assert task.kwargs == kwargs
    assert task.run() == 6


def test_create_job() -> None:
    """Test the Job class."""
    job = Job(name="test_job")

    assert isinstance(job, Job)
    assert job.name == "test_job"

    # Add a task
    job.add_task(Task(function=sum, args=([1, 2, 3],)))

    assert job.tasks[0].function == sum
    assert job.tasks[0].args == ([1, 2, 3],)
    assert job.tasks[0].kwargs == {}
    assert job.tasks[0].run() == 6
    assert len(job.tasks) == 1

    # Add another task
    job.add_task(Task(function=min, args=([5, 6, 7],)))
    assert len(job.tasks) == 2

    # Remove a task
    job.remove_task(0)
    assert len(job.tasks) == 1
    assert job.tasks[0].function == min
    assert job.tasks[0].run() == 5


def test_etl_pipeline() -> None:
    """Test the ETLPipeline base class."""
    etl_pipeline = ETLPipeline()
    assert etl_pipeline.jobs == []

    # Create a job
    job = Job(name="test_job")

    # Add a task
    job.add_task(Task(function=sum, args=([1, 2, 3],)))
    etl_pipeline.add_job(job)

    assert job in etl_pipeline.jobs
    assert len(etl_pipeline.jobs) == 1
    assert etl_pipeline.jobs[0].tasks[0].run() == 6

    # Remove a job
    etl_pipeline.remove_job(0)
    assert len(etl_pipeline.jobs) == 0



"""Event consumer. Acts as an interface between Akita and the ETL pipeline."""
from threading import Thread
from typing import Any, Protocol


class JobProducer(Protocol):
    """Job producer interface."""

    def produce_jobs(self, event: Any) -> None:
        """Produce jobs triggered by an event."""
        ...


class Queue(Protocol):
    """Queue interface."""

    @property
    def queue_list(self) -> list[Any]:
        """Return list of events in the queue."""
        ...

    @property
    def sentinel(self) -> Any:
        """Return the sentinel value."""
        ...

    def dequeue(self) -> Any:
        """Remove and return an event from the queue."""
        ...

    def peek(self) -> Any:
        """Peek first item in the queue withou removing it."""
        ...

    def get_queue_size(self) -> int:
        """Return the size of the queue."""
        ...

    def enqueue(self, event: Any) -> bool:
        """Add an event to the queue."""
        ...

    def remove(self, event: Any) -> bool:
        """Remove an event from the queue."""
        ...


class EventConsumer:
    """
    Base class for event consumers.

    Attributes
    ----------
    _queue
        Queue where events are placed.
    _job_producer
        Producer of jobs.
    _worker, optional
        Worker thread that consumes events from the and processes them to produce jobs.
    """

    def __init__(
        self, queue: Queue, job_producer: JobProducer, worker: Thread = None
    ) -> None:
        """
        Initialize the event consumer.

        Parameters
        ----------
        queue
            Queue where events are placed.
        job_producer
            Producer of jobs.
        worker, optional
            Worker thread that consumes events from the and
            processes them to produce jobs.
        """
        self._job_producer = job_producer
        self._queue = queue
        self._worker = worker

    def _is_sentinel_active(self) -> bool:
        """Return True if the sentinel is active."""
        return self._queue.sentinel()

    def _run_event_loop(self, *args, **kwargs) -> None:
        """
        Run the worker thread.

        Callback function to be used as a target by Thread.
        """
        raise NotImplementedError("Run worker not implemented.")

    def _create_worker(self, daemon: bool = True) -> Thread:
        """
        Create the worker thread.

        Parameters
        ----------
        daemon
            If `True`, runs the thread as a daemon; otherwise
            thread is not created as a daemon.

        Returns
        -------
            Worker thread that consumes events from the in-memory queue
            and processes them to produce jobs.
        """
        # Set up a worker thread to process database load
        self._worker = Thread(target=self._run_event_loop, daemon=daemon)

        return self._worker

    def run(self, daemon: bool = False) -> None:
        """Run the event consumer.

        daemon
            If `True` and the worker has not been passed to self
              runs the thread as a daemon; otherwise thread is non-daemon.
        """
        try:
            if self._worker is None:
                self._create_worker(daemon=daemon)

            self._worker.start()
        except Exception as excpt:
            self._worker.join()
            raise RuntimeError(f"Error while running event consumer: {excpt}")

        self._worker.join()



"""Test suite for the dpypeline.akita.directory_state module."""
import os
import tempfile
from typing import Generator, List

import pytest

from dpypeline.akita.directory_state import DirectoryState

FILENAMES: List[str] = ["file1.txt", "file2.txt", "file3.txt", "file4.txt"]


@pytest.fixture(scope="module")
def temp_directory_with_files() -> Generator[str, None, None]:
    """Create a temporary directory with files."""
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp()

    # Create mock files inside the temporary directory
    for filename in FILENAMES:
        file_path = os.path.join(temp_dir, filename)
        with open(file_path, "w") as file:
            file.close()

    yield temp_dir

    # Remove the temporary directory and its contents
    for root, dirs, files in os.walk(temp_dir, topdown=False):
        for file in files:  # type: ignore
            os.remove(os.path.join(root, str(file)))
        for dir in dirs:
            os.rmdir(os.path.join(root, dir))

    os.rmdir(temp_dir)


def test_get_current_directory_state(
    temp_directory_with_files: str,
    cache_dir: str,
) -> None:
    """Test the _get_directory_state method and current_state property."""
    # The test can use the prepared temporary directory
    directory_state = DirectoryState(path=temp_directory_with_files, patterns=["*.txt"])

    assert isinstance(
        directory_state, DirectoryState
    ), f"{directory_state} is not an instance of DirectoryState"

    # Test the current and stored state of the directory
    assert (
        directory_state._current_state is None
    ), f"{directory_state.current_state} != None"
    assert (
        directory_state._stored_state is None
    ), f"{directory_state.stored_state} != None"
    assert directory_state.stored_state == [], f"{directory_state.stored_state} != []"
    assert directory_state._stored_state == [], f"{directory_state.stored_state} != []"

    # Get the current state of the directory
    dir_state = directory_state._get_current_directory_state()

    # Expected results
    expected_state = []
    for filename in FILENAMES:
        expected_state.append(os.path.join(temp_directory_with_files, filename))

    assert len(dir_state) == 4
    for file in dir_state:
        assert os.path.isfile(file), f"{file} not in {dir_state}"
        assert file in expected_state, f"{file} not in {expected_state}"

    # Test the current state of the directory
    assert (
        dir_state == directory_state._current_state
    ), f"{dir_state} != {directory_state.current_state}"
    assert (
        dir_state == directory_state.current_state
    ), f"{dir_state} != {directory_state.current_state}"
    assert (
        dir_state == directory_state._current_state
    ), f"{dir_state} != {directory_state.current_state}"
    assert directory_state._stored_state == [], f"{directory_state.stored_state} != []"


def test_save_load_state(
    temp_directory_with_files: str,
    cache_dir: str,
) -> None:
    """Test the _save_state and _load_state methods and the stored_state property."""
    # The test can use the prepared temporary directory
    directory_state = DirectoryState(path=temp_directory_with_files, patterns=["*.txt"])

    assert isinstance(
        directory_state, DirectoryState
    ), f"{directory_state} is not an instance of DirectoryState."

    # Get the current state of the directory
    dir_state = directory_state._get_current_directory_state()

    # Save the state of the directory
    directory_state._save_state()

    # Assert that the state file was saved
    assert os.path.isfile(
        os.path.join(cache_dir, directory_state._state_file)
    ), f"{os.path.join(cache_dir, directory_state._state_file)} not found"

    # Load the state of the directory
    directory_state._load_state()

    assert (
        dir_state == directory_state._stored_state
    ), "State after loading is not equal to the current state of the directory."
    assert (
        dir_state == directory_state.stored_state
    ), "State after loading is not equal to the current state of the directory."

    # Create a new file in the temporary directory
    new_file_path = os.path.join(temp_directory_with_files, "file5.txt")
    with open(new_file_path, "w") as new_file:
        new_file.close()

    assert (
        dir_state == directory_state._stored_state
    ), "State is not equal to the stored state of the directory."
    assert (
        dir_state == directory_state.stored_state
    ), "State is not equal to the current state of the directory."

    assert (
        len(directory_state.stored_state) == 4
    ), "Wrong number of files in the stored state of the directory."
    assert (
        len(directory_state._stored_state) == 4
    ), "Wrong number of files in the stored state of the directory."
    assert (
        len(directory_state._current_state) == 4
    ), "Wrong number of files in the current state of the directory."

    # Test the current state of the directory
    dir_state_2 = directory_state._get_current_directory_state()
    assert (
        dir_state_2 == directory_state._current_state
    ), f"{dir_state_2} != {directory_state._current_state}"
    assert (
        dir_state_2 == directory_state.current_state
    ), f"{dir_state_2} != {directory_state.current_state}"

    # Test the stored state of the directory
    assert (
        dir_state_2 != directory_state._stored_state
    ), f"{dir_state_2} != {directory_state._current_state}"
    assert (
        dir_state_2 == directory_state.stored_state
    ), f"{dir_state_2} == {directory_state.current_state}"
    assert (
        dir_state_2 == directory_state._stored_state
    ), f"{dir_state_2} != {directory_state._current_state}"

    # Test the number of files
    assert (
        len(dir_state_2) == 5
    ), "Wrong number of files in the current state of the directory."
    assert (
        len(directory_state.current_state) == 5
    ), "Wrong number of files in the current state of the directory."
    assert (
        len(directory_state._current_state) == 5
    ), "Wrong number of files in the current state of the directory."
    assert (
        len(directory_state.stored_state) == 5
    ), "Wrong number of files in the stored state of the directory."
    assert (
        len(directory_state._stored_state) == 5
    ), "Wrong number of files in the stored state of the directory."



"""Akita, the watchdog class."""
import logging
import time
from itertools import chain
from threading import Thread
from typing import Any, Protocol

logger = logging.getLogger(__name__)


class Queue(Protocol):
    """Queue class protocol."""

    def enqueue(self, event: Any) -> bool:
        """Enqueue an event."""
        ...

    @property
    def queue_list(self) -> list[Any]:
        """Return the list of events in the queue."""
        ...

    @property
    def processed_events(self) -> list[Any]:
        """Return the list of processed events."""
        ...

    def set_sentinel_state(self, active: bool) -> bool:
        """Set the state of the end-of-queue sentinel."""
        ...


class DirectoryState(Protocol):
    """DirectoryState class protocol."""

    @property
    def stored_state(self) -> list[str]:
        """Return the stored state of the directory."""
        ...

    @property
    def current_state(self) -> list[str]:
        """Return the current state of the directory."""


class EventHandler(Protocol):
    """EventHandler class protocol."""

    def on_created(self, event: Any) -> None:
        """Return event when a file or directory is created."""
        ...

    def on_deleted(self, event: Any) -> None:
        """Return event when a file or directory is deleted."""
        ...

    def on_modified(self, event: Any) -> None:
        """Return event when a file or directory is modified."""
        ...

    def on_moved(self, event: Any) -> None:
        """Return event when a file or directory is moved."""
        ...


class Observer(Protocol):
    """Observer class protocol."""

    def start(self) -> None:
        """Start the obsever."""
        ...

    def stop(self) -> None:
        """Stop the observer."""
        ...

    def join(self) -> None:
        """Join the observer."""
        ...

    def schedule(self, event_handler: EventHandler, path: str, recursive: bool) -> None:
        """Schedules watching a path."""
        ...

    def is_alive(self) -> bool:
        """Return whether the thread is alive."""
        ...


class Akita:
    """
    Akita, the watchdog.

    The Akita watchdog holds the queue, observer, and event handler instances.
    The in-memory Queue is created upon instantiation and should be a singleton
    to provide a global point of access.

    Attributes
    ----------
    _path
        Path to watch.
    _queue
        Queue where events are placed by the event handler.
    _event_handler
        Handler responsible for matching given patterns with file paths associated
        with occurring events.
    _directory_state
        Instance that holds the state of the directory.
    _observer
        Observer thread that schedules watching directories and dispatches calls to
        event handlers.
    _event_handler
        Handler responsible for matching given patterns with file paths associated with
        occurring events.
    _monitor
        Whether or not to run the watchdog.
    _worker
        Thread that runs the watchdog.
    """

    def __init__(
        self,
        path: str,
        queue: Queue,
        event_handler: EventHandler,
        directory_state: DirectoryState,
        observer: Observer,
        monitor: bool = True,
        worker: Thread = None,
    ) -> None:
        """
        Initialize the Akita watchdog.

        Parameters
        ----------
        path
            Path to watch.
        queue
            Queue where events are placed by the event handler.
        event_handler
            Handler responsible for matching given patterns with file paths
            associated with occurring events.
        directory_state
            Instance that holds the state of the directory.
        observer
            Observer thread that schedules watching directories and dispatches
            calls to event handlers.
        monitor
            If `True`, the watchdog will monitor the given path.
        worker
            Thread that runs the watchdog.
        """
        self._path = path
        self._queue = queue
        self._observer = observer
        self._event_handler = event_handler
        self._directory_state = directory_state
        self._worker = worker
        self._monitor = monitor

    @property
    def queue(self):
        """Return the queue."""
        return self._queue

    @queue.deleter
    def queue(self):
        """Delete the queue."""
        logger.info("-" * 79)
        logger.info("Deleting in-memory queue.")
        logger.info("-" * 79)

        del self._queue

    def _get_unenqueued_files(self) -> list[str]:
        """
        Return a list of files that are not enqueued.

        Notes
        -----
        At the beggining of the run,
          Akita determines which files in the the directory being
          monitored have never been enqueued.
        This is done as follows:

        n = d - (s V q V p)

        where

        - n / unqueued_events: events not enqueued.
        - s / stored_states: events in the stored directory state.
        - d / curr_states: events in the current directory state.
        - q / queue: events in the queue when the previous session terminated.
        - p / prev_sess_states: events processed in the previous session.

        Returns
        -------
            List of files to enqueue.
        """
        stored_states = self._directory_state.stored_state
        curr_states = self._directory_state.current_state
        queue = self._queue.queue_list
        prev_sess_states = self._queue.processed_events
        unqueued_events = set(curr_states) - set(
            chain.from_iterable([queue, prev_sess_states])
        )

        logger.info("-" * 79)

        logger.info(f"Events in the stored directory state (n={len(stored_states)}):")
        for event in stored_states:
            logger.info(f"{event}")

        logger.info(f"Events in the current directory state (n={len(curr_states)}):")
        for event in curr_states:
            logger.info(f"{event}")

        logger.info(
            "Events in the queue when the previous session terminated "
            + f"(n={len(queue)}):"
        )
        for event in queue:
            logger.info(f"{event}")

        logger.info(
            "Events processed before the previous session terminated "
            f"(n={len(prev_sess_states)}):"
        )
        for event in prev_sess_states:
            logger.info(f"{event}")

        logger.info(f"Events unenqueued (n={len(unqueued_events)}):")
        for event in unqueued_events:
            logger.info(f"{event}")

        logger.info(
            f"Found {len(unqueued_events)} files not enqueued "
            + "in the current state of the directory."
        )

        logger.info("-" * 79)

        return list(unqueued_events)

    def enqueue_new_files(self) -> None:
        """Enqueue events previously unqueued."""
        # Get current state of the directory
        # TODO: Check if delimiter has to be "\" when running on Windows.
        not_enqueued_state = sorted(
            self._get_unenqueued_files(), key=lambda f: f.rsplit("/", 1)[-1]
        )

        for event in not_enqueued_state:
            self._queue.enqueue(event)

    def _run_watchdog(self) -> None:
        """Run the watchdog."""
        logger.debug("Starting the watchdog...")
        self._observer.schedule(self._event_handler, self._path, recursive=True)
        self._observer.start()

        try:
            while self._observer.is_alive():
                time.sleep(1)
        finally:
            self._observer.stop()
            self._observer.join()

    def _create_worker(self, daemon: bool = False) -> Thread:
        """
        Create the worker thread.

        Parameters
        ----------
        daemon
            If `True`, runs the thread as a daemon; otherwise
            thread is not created as a daemon.

        Returns
        -------
            Worker thread that consumes events from the in-memory queue
            and processes them to produce jobs.
        """
        # Set up a worker thread to process database load
        self._worker = Thread(target=self._run_watchdog, daemon=daemon)

        return self._worker

    def _stop_worker(self, max_attempts=10) -> None:
        """
        Stop the worker thread.

        Parameters
        ----------
        max_attempts
            Maximum number of attempts to stop the worker thread.
        """
        while self._worker.is_alive() and max_attempts > 0:
            logger.debug(
                f"{max_attempts} attempts remaining to stop the worker thread."
            )
            if self._observer.is_alive():
                logger.debug("Stopping the watchdog...")
                self._observer.stop()
                self._observer.join()
                logger.debug("Watchdog has stopped successfully.")

            logger.debug("Attempting to stop the worker thread...")
            self._worker.join(timeout=1)
            max_attempts -= 1

        if max_attempts == 0:
            raise RuntimeError("Failed to stop the worker thread.")
        else:
            logger.debug("Worker thread has stopped successfully.")

    def stop(self, max_attempts=10) -> None:
        """
        Stop the observer and worker thread.

        Parameters
        ----------
        max_attempts
            Maximum number of attempts to stop the worker thread.
        """
        if self._worker is not None:
            logger.info("Stopping Akita...")
            self._stop_worker(max_attempts)
            logger.info("Akita has stopped successfully.")
        else:
            logger.info("Attempted to stop akita but no worker thread has been found.")

    def run(
        self, monitor: bool = True, enqueue_new_files: bool = True, daemon: bool = False
    ) -> None:
        """
        Run the Akita watchdog.

        Parameters
        ----------
        monitor
            If `True` monitors runs the watchdog. `False` otherwise.
        enqueue_new_files
            If `True`, enqueues events previously unqueued.
        daemon
            If `True`, runs the thread as a daemon.
        """
        if enqueue_new_files:
            # Add new files found in the directory since the last
            # time the watchdog was run.
            self.enqueue_new_files()

        if self._monitor:
            logger.info("Starting Akita in monitor mode...")
            self._queue.set_sentinel_state(active=False)

            try:
                if self._worker is None:
                    self._create_worker(daemon=daemon)

                self._worker.start()
            except Exception as excpt:
                self._worker.join()
                raise RuntimeError(f"Error while running the watchdog: {excpt}")
        else:
            self._queue.set_sentinel_state(active=True)
